use std::borrow::Cow;

#[derive(Clone, Debug, PartialEq, Eq, Hash)]
pub struct ItemPath<'a> {
    items: Vec<Cow<'a, str>>,
}

#[derive(Clone, Debug, PartialEq, Eq, Hash)]
pub struct OwnedItemPath {
    items: Vec<String>,
}

impl OwnedItemPath {
    pub fn name(&self) -> Option<String> {
        self.items.last().cloned()
    }
}

impl From<ItemPath<'_>> for OwnedItemPath {
    fn from(value: ItemPath<'_>) -> Self {
        Self {
            items: value
                .items
                .into_iter()
                .map(|v| v.into_owned())
                .collect::<Vec<String>>(),
        }
    }
}

impl<'a> ItemPath<'a> {
    pub fn new() -> Self {
        Self { items: Vec::new() }
    }

    pub fn from_slices(slice: &[&'a str]) -> Self {
        Self {
            items: slice.into_iter().map(|i| (*i).into()).collect::<Vec<_>>(),
        }
    }

    pub fn push(mut self, item: impl Into<Cow<'a, str>>) -> Self {
        self.items.push(item.into());
        self
    }

    pub fn name(&self) -> Option<String> {
        self.items.last().cloned().map(|item| item.into_owned())
    }
}

impl<'a> TryFrom<Pattern<'a>> for ItemPath<'a> {
    type Error = ();

    fn try_from(value: Pattern<'a>) -> Result<Self, Self::Error> {
        let mut items = Vec::new();

        for i in value.items {
            match i {
                PatternEl::Name(n) => items.push(n),
                PatternEl::Glob => return Err(()),
            }
        }

        Ok(Self { items })
    }
}

#[derive(Clone, Debug)]
pub struct Pattern<'a> {
    items: Vec<PatternEl<'a>>,
}

impl<'a> Pattern<'a> {
    pub fn new() -> Self {
        Self { items: Vec::new() }
    }

    pub fn add_name(mut self, name: &'a str) -> Self {
        self.items.push(PatternEl::Name(name.into()));
        self
    }

    pub fn add_glob(mut self) -> Self {
        self.items.push(PatternEl::Glob);
        self
    }
}

#[derive(Clone, Debug)]
enum PatternEl<'a> {
    Name(Cow<'a, str>),
    Glob,
}
