use std::path::{Path, PathBuf};

use rowan::ast::{AstNode, AstPtr};

use crate::lst_parser::{
    grammar::source_file,
    input::Input,
    output::Output,
    syntax_kind::{self, SyntaxKind},
    Parser,
};

use super::{
    error::Error,
    nodes::{Def, Mod, Root, Use},
};

#[derive(Debug)]
pub struct SourceFile {
    lst: Root,
    modules: Vec<AstPtr<Mod>>,
    defs: Vec<AstPtr<Def>>,
    uses: Vec<AstPtr<Use>>,
}

impl SourceFile {
    pub fn parse_from(path: &Path, source_text: String) -> (Self, Vec<Error>) {
        let toks = syntax_kind::lex(&source_text);
        let input = Input::new(&toks);
        let mut parser = Parser::new(input);

        source_file(&mut parser);

        let parser_out = parser.finish();
        let out = Output::from_parser_output(toks, parser_out);

        let lst = out.syntax();

        let mut modules = Vec::new();
        let mut defs = Vec::new();
        let mut uses = Vec::new();

        for c in lst.children() {
            match c.kind() {
                SyntaxKind::MODULE => modules.push(AstPtr::new(&Mod::cast(c).unwrap())),
                SyntaxKind::DEF => defs.push(AstPtr::new(&Def::cast(c).unwrap())),
                SyntaxKind::USE => uses.push(AstPtr::new(&Use::cast(c).unwrap())),
                _ => {}
            }
        }

        (
            Self {
                lst: Root::cast(lst).unwrap(),
                modules,
                defs,
                uses,
            },
            out.errors()
                .into_iter()
                .map(|err| Error::from_syntax(path.to_owned(), err))
                .collect(),
        )
    }

    pub fn tree(&self) -> &Root {
        &self.lst
    }

    pub fn top_level_modules(&self) -> Vec<AstPtr<Mod>> {
        self.modules.clone()
    }

    pub fn top_level_defs(&self) -> Vec<AstPtr<Def>> {
        self.defs.clone()
    }

    pub fn top_level_uses(&self) -> Vec<AstPtr<Use>> {
        self.uses.clone()
    }
}
